@@ -30,14 +30,10 @@
 #include "fcitx-gclient/fcitxgclient.h"
 #include "fcitx-gclient/fcitxgwatcher.h"
 #include "fcitximcontext.h"
+#include "fcitxtheme.h"
+#include "gtk3inputwindow.h"
 
-#if !GTK_CHECK_VERSION(2, 91, 0)
-#define DEPRECATED_GDK_KEYSYMS 1
-#endif
-
-#if GTK_CHECK_VERSION(2, 24, 0)
-#define NEW_GDK_WINDOW_GET_DISPLAY
-#endif
+using namespace fcitx::gtk;
 
 static constexpr uint32_t HandledMask = (1 << 24);
 static constexpr uint32_t IgnoredMask = (1 << 25);
@@ -102,6 +98,8 @@ struct _FcitxIMContext {
     struct xkb_compose_state *xkbComposeState;
 
     GdkEvent *gdk_event;
+
+    Gtk3InputWindow *candidate_window;
 };
 
 struct _FcitxIMContextClass {
@@ -212,6 +210,7 @@ static guint _key_snooper_id = 0;
 static FcitxGWatcher *_watcher = NULL;
 static struct xkb_context *xkbContext = NULL;
 static struct xkb_compose_table *xkbComposeTable = NULL;
+static ClassicUIConfig *_uiconfig = nullptr;
 
 void fcitx_im_context_register_type(GTypeModule *type_module) {
     static const GTypeInfo fcitx_im_context_info = {
@@ -406,6 +405,7 @@ static void fcitx_im_context_init(FcitxIMContext *context, gpointer) {
     static gsize has_info = 0;
     if (g_once_init_enter(&has_info)) {
         _watcher = fcitx_g_watcher_new();
+        _uiconfig = new ClassicUIConfig;
         fcitx_g_watcher_set_watch_portal(_watcher, TRUE);
         fcitx_g_watcher_watch(_watcher);
         g_object_ref_sink(_watcher);
@@ -437,13 +437,9 @@ static void fcitx_im_context_init(FcitxIMContext *context, gpointer) {
     if (context->is_wayland) {
         fcitx_g_client_set_display(context->client, "wayland:");
     } else {
-#if GTK_CHECK_VERSION(3, 0, 0)
         if (GDK_IS_X11_DISPLAY(gdk_display_get_default())) {
-#endif
             fcitx_g_client_set_display(context->client, "x11:");
-#if GTK_CHECK_VERSION(3, 0, 0)
         }
-#endif
     }
     g_signal_connect(context->client, "connected",
                      G_CALLBACK(_fcitx_im_context_connect_cb), context);
@@ -466,6 +462,8 @@ static void fcitx_im_context_init(FcitxIMContext *context, gpointer) {
 
 static void fcitx_im_context_finalize(GObject *obj) {
     FcitxIMContext *context = FCITX_IM_CONTEXT(obj);
+    delete context->candidate_window;
+    context->candidate_window = nullptr;
 
     fcitx_im_context_set_client_window(GTK_IM_CONTEXT(context), NULL);
 
@@ -493,11 +491,21 @@ static void fcitx_im_context_finalize(GObject *obj) {
 static void fcitx_im_context_set_client_window(GtkIMContext *context,
                                                GdkWindow *client_window) {
     FcitxIMContext *fcitxcontext = FCITX_IM_CONTEXT(context);
+    if (client_window != fcitxcontext->client_window) {
+        delete fcitxcontext->candidate_window;
+        fcitxcontext->candidate_window = nullptr;
+    }
     if (!client_window)
         return;
 
     g_clear_object(&fcitxcontext->client_window);
     fcitxcontext->client_window = GDK_WINDOW(g_object_ref(client_window));
+    if (!fcitxcontext->candidate_window) {
+        fcitxcontext->candidate_window =
+            new Gtk3InputWindow(_uiconfig, fcitxcontext->client);
+        fcitxcontext->candidate_window->setParent(fcitxcontext->client_window);
+        fcitxcontext->candidate_window->setCursorRect(fcitxcontext->area);
+    }
 }
 
 static gboolean
@@ -669,7 +677,6 @@ static void _fcitx_im_context_update_preedit(FcitxIMContext *context,
                                          (gpointer *)&widget);
                 if (GTK_IS_WIDGET(widget)) {
                     hasColor = true;
-#if GTK_CHECK_VERSION(3, 0, 0)
                     GtkStyleContext *styleContext =
                         gtk_widget_get_style_context(widget);
                     GdkRGBA fg_rgba, bg_rgba;
@@ -692,11 +699,6 @@ static void _fcitx_im_context_update_preedit(FcitxIMContext *context,
                             CLAMP((gint)(bg_rgba.green * 65535), 0, 65535);
                         bg.blue = CLAMP((gint)(bg_rgba.blue * 65535), 0, 65535);
                     }
-#else
-                    GtkStyle *style = gtk_widget_get_style(widget);
-                    fg = style->text[GTK_STATE_SELECTED];
-                    bg = style->base[GTK_STATE_SELECTED];
-#endif
                 }
             }
 
@@ -870,6 +872,9 @@ static void fcitx_im_context_set_cursor_location(GtkIMContext *context,
         return;
     }
     fcitxcontext->area = *area;
+    if (fcitxcontext->candidate_window) {
+        fcitxcontext->candidate_window->setCursorRect(fcitxcontext->area);
+    }
 
     if (fcitx_g_client_is_valid(fcitxcontext->client)) {
         _set_cursor_location_internal(fcitxcontext);
@@ -906,28 +911,11 @@ static gboolean _set_cursor_location_internal(FcitxIMContext *fcitxcontext) {
     {
         if (area.x == -1 && area.y == -1 && area.width == 0 &&
             area.height == 0) {
-#if GTK_CHECK_VERSION(2, 91, 0)
             area.x = 0;
             area.y += gdk_window_get_height(fcitxcontext->client_window);
-#else
-            gint w, h;
-            gdk_drawable_get_size(fcitxcontext->client_window, &w, &h);
-            area.y += h;
-            area.x = 0;
-#endif
         }
-
-#if GTK_CHECK_VERSION(2, 18, 0)
         gdk_window_get_root_coords(fcitxcontext->client_window, area.x, area.y,
                                    &area.x, &area.y);
-#else
-        {
-            int rootx, rooty;
-            gdk_window_get_origin(fcitxcontext->client_window, &rootx, &rooty);
-            area.x += rootx;
-            area.y += rooty;
-        }
-#endif
     }
     int scale = 1;
 #if GTK_CHECK_VERSION(3, 10, 0)
@@ -1086,6 +1074,7 @@ void _fcitx_im_context_set_capability(FcitxIMContext *fcitxcontext,
         }
         if (fcitxcontext->is_wayland) {
             flags |= (guint64)fcitx::FcitxCapabilityFlag_RelativeRect;
+            flags |= (guint64)fcitx::FcitxCapabilityFlag_ClientSideInputPanel;
         }
         flags |= (guint64)fcitx::FcitxCapabilityFlag_KeyEventOrderFix;
         flags |= (guint64)fcitx::FcitxCapabilityFlag_ReportKeyRepeat;
@@ -1288,11 +1277,7 @@ static GdkEventKey *_create_gdk_event(FcitxIMContext *fcitxcontext,
     event->length = 0;
     event->hardware_keycode = 0;
     if (event->window) {
-#ifndef NEW_GDK_WINDOW_GET_DISPLAY
-        GdkDisplay *display = gdk_display_get_default();
-#else
         GdkDisplay *display = gdk_window_get_display(event->window);
-#endif
         GdkKeymap *keymap = gdk_keymap_get_for_display(display);
         GdkKeymapKey *keys;
         gint n_keys = 0;
@@ -1343,19 +1328,10 @@ static GdkEventKey *_create_gdk_event(FcitxIMContext *fcitxcontext,
             g_locale_from_utf8(buf, len, NULL, &bytes_written, NULL);
         if (event->string)
             event->length = bytes_written;
-#ifdef DEPRECATED_GDK_KEYSYMS
-    } else if (keyval == GDK_Escape) {
-#else
     } else if (keyval == GDK_KEY_Escape) {
-#endif
         event->length = 1;
         event->string = g_strdup("\033");
-    }
-#ifdef DEPRECATED_GDK_KEYSYMS
-    else if (keyval == GDK_Return || keyval == GDK_KP_Enter) {
-#else
-    else if (keyval == GDK_KEY_Return || keyval == GDK_KEY_KP_Enter) {
-#endif
+    } else if (keyval == GDK_KEY_Return || keyval == GDK_KEY_KP_Enter) {
         event->length = 1;
         event->string = g_strdup("\r");
     }
@@ -1364,7 +1340,6 @@ static GdkEventKey *_create_gdk_event(FcitxIMContext *fcitxcontext,
         event->length = 0;
         event->string = g_strdup("");
     }
-#if GTK_CHECK_VERSION(3, 0, 0)
     // Set the event device to be the same device.
     if (fcitxcontext->gdk_event) {
         gdk_event_set_device((GdkEvent *)event,
@@ -1373,7 +1348,6 @@ static GdkEventKey *_create_gdk_event(FcitxIMContext *fcitxcontext,
             (GdkEvent *)event,
             gdk_event_get_source_device(fcitxcontext->gdk_event));
     }
-#endif
 out:
     return event;
 }
@@ -1466,32 +1440,15 @@ void _fcitx_im_context_connect_cb(FcitxGClient *im, void *user_data) {
     FcitxIMContext *context = FCITX_IM_CONTEXT(user_data);
     Display *display = NULL;
     if (context->client_window) {
-#if GTK_CHECK_VERSION(3, 0, 0)
         if (GDK_IS_X11_WINDOW(context->client_window)) {
             display = GDK_WINDOW_XDISPLAY(context->client_window);
         }
-#else
-        if (GDK_IS_WINDOW(context->client_window)) {
-            auto gdkDisplay = gdk_window_get_display(context->client_window);
-            if (gdkDisplay) {
-                auto x11DisplayType = g_type_from_name("GdkDisplayX11");
-                if (x11DisplayType &&
-                    G_TYPE_CHECK_INSTANCE_TYPE(gdkDisplay, x11DisplayType)) {
-                    display = GDK_DISPLAY_XDISPLAY(gdkDisplay);
-                }
-            }
-        }
-#endif
     }
     if (!display) {
         GdkDisplay *gdkDisplay = gdk_display_get_default();
-#if GTK_CHECK_VERSION(3, 0, 0)
         if (GDK_IS_X11_DISPLAY(gdkDisplay)) {
-#endif
             display = GDK_DISPLAY_XDISPLAY(gdkDisplay);
-#if GTK_CHECK_VERSION(3, 0, 0)
         }
-#endif
     }
 
     if (display) {
